/* nylon - (socks{4,5}) proxy
 * http://mesh.eecs.umich.edu/projects/nylon
 * marius aamodt eriksen (marius@umich.edu) */

#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <errno.h>
#include "mnet.h"
#include <sys/ioctl.h>
#include <net/if.h>
#include "cfg.h"
#include "util.h"

#if defined(__sun__)
#include <sys/sockio.h>
#endif /* __sun__ */

#define DEFAULT_MALLOC 8   /* userids usually aren't that large */
#define IFMAX IFNAMSIZ
#define S4_MAX_USERID 256

void usage(char**);
void fatal_error(char*);
void die(char*);
void printd(char*);
void wait_for_connections(int);
void negotiate_connection(int, int);
void negotiate4(int, int);
void negotiate5(int, int);
void establish_connection(int, int, int);
void sigchld(int);
void sigterm(int);
struct in_addr get_ip_from_if(char*);

typedef struct {
	unsigned char  vn,
		           cd;
	unsigned short dport;
	unsigned long  dip;
	unsigned char  terminator;
} socks4_hdr;

typedef struct {
	unsigned char	  vn,
		              cd,
		             rsv,
		            atyp;
	unsigned long  daddr;
	unsigned short dport;
} socks5_req;

/* version reply */
typedef struct {
	unsigned char ver,
		          res;
} socks5_vrepl;  

int verbose, no_connections, log, active_connections;
char cvsid[] = "$Id: nylon.c,v 1.9 2001/06/17 07:09:34 beriksen Exp $";

char *cinf    = NULL;
char *inf     = NULL;
char *ipallow = NULL;
char *ipdeny  = NULL;
char *pidfile = "/tmp/nylon.pid";


/* main() - read in command line options and act upon them, then
 * fork-and-exit for ``daemon like'' behavior */
int main (int argc, char **argv)
{
	int opt, port = 1080, bgproc;
	char *_inf, *_cinf, apid[6];
	FILE *pidfd;

	/* XXX - apid, not portable */

	active_connections = log = verbose  = 0;
	no_connections = 1;

	/* take care of configuration */

	conf_init();
	CONF_SAVE(inf,  conf_get_str("Server", "Binding-Interface"));
	CONF_SAVE(cinf, conf_get_str("Server", "Connecting-Interface"));
	CONF_SAVE(pidfile, conf_get_str("General", "PIDFile"));	
	port           = conf_get_num("Server", "Port", 1080);
	no_connections = conf_get_num("General", "No-Simultaneous-Conn", 10);
	log            = conf_get_num("General", "Log", 0);
	verbose        = conf_get_num("General", "Verbose", 0);		

	while ((opt = getopt(argc, argv, "hvVp:n:i:I:")) != -1) 
		switch (opt) {
		case 'h':
			usage(argv);
		case 'v':
			verbose = 1;
			break;
		case 'V':
			printf("nylon version " VERSION ":\n%s\n", cvsid);
			exit(0);
		case 'p':
			port = atoi(optarg);
			break;
		case 'n':
			no_connections = atoi(optarg);
			break;
		case 'i':
			_inf = (char*)malloc(strlen(optarg)+1);
			strlcpy(_inf, optarg, strlen(optarg)+1);
			inf = _inf;
			if (verbose)
				printf("binding with %s\n", inf);
			break;
		case 'I':
			_cinf = (char*)malloc(strlen(optarg)+1);
			strlcpy(_cinf, optarg, strlen(optarg)+1);
			cinf = _cinf;
			if (verbose)
				printf("binding with %s\n", cinf);			
			break;
		default:
			break;
		}

	signal(SIGCHLD, sigchld);    /* get 'em zombies away! */
	signal(SIGTERM, sigterm);    /* graceful exit         */

	bgproc = fork();

	if (bgproc == -1) 
		fatal_error("fork() failed");
	if (bgproc != 0) {
		if (verbose)
			printf("nylon launched with pid %i\n", bgproc);
		if ( !(pidfd = fopen(pidfile, "w")) )
			fatal_error("could not write to pid file");
		snprintf(apid, 6, "%i", bgproc);
		fputs(apid, pidfd);
		fclose(pidfd);
		exit(0);
	}
	/* we are now the child */
	wait_for_connections(port);
	/* NOTREACHED */
	exit(0);
}

void wait_for_connections(int portno)
{
	struct sockaddr_in serv_addr, cli_addr;
	int sfd, nsfd, nproc, session_no = 0;
	socklen_t cli_len = sizeof(struct sockaddr);
	char ifip[128];
	
	bzero(&serv_addr, sizeof(serv_addr));
	bzero(&cli_addr , sizeof(cli_addr));

	if ((sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		fatal_error("socket() failed");

	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port   = htons(portno);
	if (*inf) 
		serv_addr.sin_addr = get_ip_from_if(inf);
	else
		serv_addr.sin_addr.s_addr = INADDR_ANY;
		
	if (bind(sfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
		fatal_error("bind() failed");

	if (listen(sfd, no_connections) == -1)
		fatal_error("listen() failed");

	if (verbose)
		printf("starting to listen...\n");
	for (;;) {
		if ((nsfd = accept(sfd, (struct sockaddr*) &cli_addr, &cli_len)) == -1 )
			fatal_error("accept() failed");

		if (verbose)
			printf("accept()ed\n");

		if (active_connections > no_connections-1)
			continue;

		session_no++;
		nproc = fork();
		if (nproc == -1)
			fatal_error("fork() failed");
		if (nproc == 0) {
			if (verbose)
				printf("connect! starting negotiations\n");
			negotiate_connection(nsfd, session_no);
			exit(0);
		}
		close(nsfd);
		active_connections++;
	}
}

/* multiplex depending on version number */
void negotiate_connection(int sfd, int session_no)
{
	unsigned char vn;

	if ( recvn(sfd, &vn, 1, 0) == -1 )
		fatal_error("error receiving version number\n");

	if (vn == 4)               /* SOCKS4 */
		negotiate4(sfd, session_no);
	else if (vn == 5)          /* SOCKS5 */
		negotiate5(sfd, session_no);
	else
		die("No support for requested protocol version\n");
}

void negotiate4(int sfd, int session_no)
{
	unsigned char vn, cd, *userid, *methods, method;
	unsigned short dport;
	unsigned long dip;
	unsigned i = 0, j = 1, nmethods;
	struct sockaddr_in target_addr, connect_addr;
	int target_sfd;
	socks4_hdr hdr4;

	if (verbose)
		printf("entering SOCKS4 specific negotiation");

	userid = (unsigned char*)malloc(DEFAULT_MALLOC);

	hdr4.vn = 4;
	if ( recvn(sfd, &hdr4.cd, 7, 0) == -1 )
		fatal_error("error receiving header\n");

	if (verbose)
		printf("getting user id\n");

	while (recvn(sfd, userid+i, 1, 0) != -1)
		if ((userid[i]) == 0) {
			userid[i] = '\0';
			hdr4.terminator = 0;
			break;
		}
		else {
			if ((++i)%DEFAULT_MALLOC == 0)
				userid = (unsigned char*)realloc(userid, (++j)*DEFAULT_MALLOC);
			if ((j*DEFAULT_MALLOC) > S4_MAX_USERID)
				fatal_error("header too long, bailing out");
		}

	
	if (hdr4.cd != 1)
		die("SOCKS4: unsupported command!\n");

	if ( (target_sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		fatal_error("could not create socket");

	bzero(&target_addr, sizeof(target_addr));
	target_addr.sin_family      = AF_INET;
	target_addr.sin_port        = hdr4.dport;   /* already in network order */
	target_addr.sin_addr.s_addr = hdr4.dip;     /* ditto                    */

	if (*cinf) {
		connect_addr.sin_addr = get_ip_from_if(cinf);
		connect_addr.sin_port = 0;
		if ( bind(target_sfd, (struct sockaddr*) &connect_addr,
				  sizeof(connect_addr)) < 0 )
			fatal_error("bind failed");
	}
		
	if (verbose)
		printf("trying to connec to target host...");

	if (connect(target_sfd,
				(struct sockaddr*)&target_addr, sizeof(target_addr)) < 0) 
		hdr4.cd = 91;
	else
		hdr4.cd = 90;

	if ( sendn(sfd, &hdr4, 9, 0) == -1 )
		fatal_error("error replying to initiating host");

	if (cd == 91)
		fatal_error("error connecting to target host");
	
	if (verbose)
		printf("connected!\n");

	establish_connection(sfd, target_sfd, session_no);
	
	return;
}


void negotiate5(int sfd, int session_no)
{
	unsigned i = 0, j = 1, reqlen = sizeof(socks5_req);
	struct sockaddr_in target_addr, connect_addr;
	int target_sfd;
	socks5_req req5;
	socks5_vrepl rep5;
	unsigned char nmethods, *methods, method;
	
	if (verbose)
		printf("initiating SOCKS5 connection\n");

	method  = 0;
	req5.vn = 5;

	if ( recvn(sfd, &nmethods, 1, 0) == -1 )
		fatal_error("error in receiving header");
	if (verbose)
		printf("got methodn == %i\n", nmethods);
	methods = (unsigned char*)malloc(nmethods);
	if ( recvn(sfd, methods, nmethods, 0) == -1 )
		fatal_error("error in receiving header");
	if (verbose)
		printf("got methods\n");
	/* we don't support any authentication methods yet, so simply
	 * ignore it and send reply with no authentication required. */
	if (verbose) {
		printf("they support method: %i\n", *methods);
		printf("sending reply: %i | %i\n", req5.vn, method);
	}

	free(methods);

	rep5.ver = 5;
	rep5.res = 0;

	if ( sendn(sfd, &rep5, 2, 0) == -1)
		fatal_error("error in sending reply");

	if (verbose)
		printf("sent no auth needed\n");

	if ( recvn(sfd, &req5, 4, 0) == -1 )
		fatal_error("Error receiving SOCKS5 request");
	if (req5.vn != 5)
		die("Error in header");
	if (req5.cd != 1)
		die("Unsupported SOCKS5 request");

	if (verbose)
		printf("request == %i\n", req5.cd);

	bzero(&target_addr, sizeof(target_addr));

	if (req5.atyp == 1) {   /* IPV4 Addr */
		if (verbose)
			printf("they want ipv4 addr\n");
		if (recvn(sfd, &req5.daddr, 4, 0) == -1)
			fatal_error("Error receiving request");
		target_addr.sin_family      = AF_INET;
		target_addr.sin_addr.s_addr = req5.daddr;

    } else if (req5.atyp == 3) {  /* domain name */
		unsigned char len;
		char *hostname;
		struct hostent *hent;
		if (recvn(sfd, &len, 1, 0) == -1)
			fatal_error("error receiving request");
		if (verbose)
			printf("domain name %i characters long\n", len);
		hostname = (char*)malloc(len);
		if (recvn(sfd, &hostname, len, 0) == -1)
			fatal_error("error receiving hostname");
		hostname[len] = '\0';
		hent = gethostbyname(hostname);
		target_addr.sin_family      = AF_INET;
		target_addr.sin_addr.s_addr = htonl(*hent->h_addr);
		free(hostname);
	} else 
		die("Unsupported address type");

	if (recvn(sfd, &req5.dport, 2, 0) == -1)
		fatal_error("Error receving request");
	target_addr.sin_port = req5.dport;

	if (verbose) {
		printf("target host:\n");
		printf("ip: %s\n", inet_ntoa(target_addr.sin_addr));
		printf("p : %i\n", ntohs(req5.dport));
	}

	if ( (target_sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		fatal_error("could not create socket");

	if (verbose)
		printf("making connection...\n");

	if (*cinf) {
		connect_addr.sin_addr = get_ip_from_if(cinf);
		connect_addr.sin_port = 0;
		if ( bind(target_sfd, (struct sockaddr*) &connect_addr,
				  sizeof(connect_addr)) < 0 )
			fatal_error("bind failed");
	}



	if (connect(target_sfd,
				(struct sockaddr*)&target_addr, sizeof(target_addr)) < 0) {
		if (verbose)
			printf("unsuccessful connect!\n");
		req5.cd = 1;
	} else {
		if (verbose)
			printf("successful connect!\n");
		req5.cd = 0;
	}

	if (verbose)
		printf("made connection...\n");
	
	if ( sendn(sfd, &req5, 10, 0) == -1 )
		fatal_error("error replying to initiating host");

	if (req5.cd == 1)
		fatal_error("error connecting to target host");
	
	establish_connection(sfd, target_sfd, session_no);
	
	return;
}


void establish_connection(int client, int target, int session_no)
{
	fd_set rfds;
	int nfds;
	struct timeval tmout;
	unsigned char buf[1024];
	unsigned bcount;


	nfds = (client > target?client:target)+1;
	
	for (;;) {
		tmout.tv_usec = 0; tmout.tv_sec = 60;
		FD_ZERO(&rfds);
		FD_SET(client, &rfds); FD_SET(target, &rfds);
		select(nfds, &rfds, NULL, NULL, &tmout);
		bzero(buf, 1024);
		if (FD_ISSET(client, &rfds)) {
			if ( (bcount = recv(client, buf, 1024, 0)) == -1 )
				fatal_error("error in recv()");
			if (bcount == 0)
				if (verbose)
					die("client terminated connection");
				else
					exit(0);
			if (verbose)
				printf("session %i: %i bytes client->target\n", session_no,
					   bcount);
			if ( send(target, buf, bcount, 0) == -1 )
				fatal_error("error in send()");
		} else if (FD_ISSET(target, &rfds)) {
			if ( (bcount = recv(target, buf, 1024, 0)) == -1 )
				fatal_error("error in recv()");
			if (bcount == 0)
				if (verbose)
					die("server terminated connection");
				else
					exit(0);
			if (verbose)
				printf("session %i: %i bytes target->client\n", session_no,
					   bcount);
			if ( send(client, buf, bcount, 0) == -1 )
				fatal_error("error in send()");			
		} else
			fatal_error("select() failed");
	}
}

struct in_addr get_ip_from_if(char *inf)
{
	struct ifreq ifr;
	struct sockaddr_in *sin;
	int bogusfd;
	char ifip[14];
	
	strlcpy(ifr.ifr_name, inf, IFMAX);
	bogusfd = socket(AF_INET, SOCK_STREAM, 0);
	ioctl(bogusfd, SIOCGIFADDR, &ifr);
	sin = (struct sockaddr_in *) &ifr.ifr_addr;
	/* display... */
	if (verbose) {
		inet_ntop(AF_INET, &sin->sin_addr, ifip, sizeof(ifip));
		printf("binding to address %s\n", ifip);
	}
	return sin->sin_addr;
}

void sigterm (int sign)
{
	unlink(pidfile);
	exit(0);
}

void sigchld (int sign)
{
	pid_t pid;
	int   stat;

	pid = wait(&stat);
	active_connections--;
	if (verbose)
		printf("decremented active count, now %i\n", active_connections);
	return;
}

void usage (char **argv)
{
	printf("mproxy version " VERSION " usage:\n"
		   "%s -hvV [-p portno] [-i if] [-I if]\n"
		   "   h - help (this)\n"
		   "   v - verbose\n"
		   "   V - version\n"
		   "   i - bind to specified interface (listening)\n"
		   "   I - bind to specified interface (connecting)\n"		   		   
		   "   p - port number\n", *argv);
	exit(0);
}

void fatal_error(char *msg)
{
	fprintf(stderr, "fatal error: %s\n", msg);
	if (errno)
		perror("");
	exit(1);
}

void die(char *msg)
{
	printf("%s\n", msg);
	exit(0);
}

void printd(char *msg)
{
	fprintf(stderr, "%s\n", msg);
}



/* copyright (c) 2001  
 * marius aamodt eriksen
 * all rights reserved
 * 
 * permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any purpose,
 * so long as the name of the author is not used in any advertising or 
 * publicity pertaining to the use or distribution of this software 
 * without specific, written prior authorization.  if the above copyright 
 * notice or any other identification of the author is included in any 
 * copy of any portion of this software, then the disclaimer below must 
 * also be included.
 * 
 * this software is provided as is, without representation from the
 * the author as to its fitness for any purpose, and withoutwarranty by 
 * the author of any kind, either express or implied, including without 
 * limitation the implied warranties of merchantability and fitness for 
 * a particular purpose.  the author shall not be liable for any damages,
 * including special, indirect, incidental, or consequential damages,
 * with respect to any claim arising out or in connection with the use
 * of the software, even if it has been or is hereafter advised of the
 * possibility of such damages. */
